# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Optional
import numpy as np
import pandas as pd

from rarity.data_loader import CSVDataLoader, DataframeLoader
from rarity.interpreters.structured_data.base_interpreters import BaseInterpreters
from rarity.utils.methods import calculate_kl_div, get_optimum_bin_size
from rarity.utils.common_functions import insert_index_col, is_regression, is_classification


class IntFeatureDistribution(BaseInterpreters):
    '''
    Transform raw data into input format suitable for visualization on feature distribution

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module
    '''
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader]):
        super().__init__(data_loader)
        self.df_features = insert_index_col(self.data_loader.get_features())
        # default range set as last 20% of dataset sample size
        self.df_default_range = self.df_features.iloc[np.r_[int(len(self.df_features) * 0.8):, :]]
        self.features = self.df_features.columns

    def _get_df_sliced(self, start_idx: int, stop_idx: int):
        '''
        Slice dataframe to the specific range.
        '''
        if start_idx is not None and stop_idx is None:
            df_sliced = self.df_features.iloc[np.r_[start_idx:, :]]
        elif start_idx is None and stop_idx is not None:
            df_sliced = self.df_features.iloc[np.r_[:stop_idx, :]]
        elif start_idx is not None and stop_idx is not None:
            df_sliced = self.df_features.iloc[np.r_[start_idx:stop_idx, :]]
        else:  # range is not specified
            df_sliced = self.df_features
            if is_regression(self.analysis_type):
                df_sliced = self.df_default_range
        return df_sliced

    def _get_single_feature_df_with_binning(self, df: pd.DataFrame, feature: str):
        '''
        For regression task only.
        Function to find optimum bin-size on sliced df for distribution comparison
        '''
        df_viz_specific_feat = df[['dataset_type', feature]]
        optimum_bin_size = get_optimum_bin_size(df_viz_specific_feat[feature])
        df_viz_specific_feat['bin_group'] = pd.cut(df_viz_specific_feat[feature], optimum_bin_size, labels=list(range(optimum_bin_size)))
        return df_viz_specific_feat, optimum_bin_size

    def _get_probabilities_by_bin_group(self, df_viz: pd.DataFrame, bin_count: int):
        '''
        For regression task only.
        Function to tap-out customized df for ease of getting probabilities based on bin group for reference df and sliced df
        '''
        def _interim_df_xformed_from_bin_group(specific_pd_series: pd.Series, col_name: str):
            interim_df = specific_pd_series.value_counts().rename_axis('bin_group').reset_index(name=col_name).sort_values('bin_group')
            return interim_df

        df_probs_by_bin = pd.DataFrame()
        df_probs_by_bin['bin_group'] = [n for n in range(bin_count)]
        df_filter_ref = df_viz.loc[lambda x:x['dataset_type'] == 'df_reference', :]
        df_filter_sliced = df_viz.loc[lambda x:x['dataset_type'] == 'df_sliced', :]

        df_temp_ref = _interim_df_xformed_from_bin_group(df_filter_ref['bin_group'], 'df_ref_counts')
        df_temp_sliced = _interim_df_xformed_from_bin_group(df_filter_sliced['bin_group'], 'df_sliced_counts')

        df_kl_div = df_probs_by_bin.merge(df_temp_ref, how='left', on='bin_group')
        df_kl_div = df_kl_div.merge(df_temp_sliced, how='left', on='bin_group')
        df_kl_div.fillna(0, inplace=True)
        df_kl_div['df_ref_counts'] = df_kl_div['df_ref_counts'].replace(0, 1e-15)  # to avoid division by zero
        df_kl_div['df_sliced_counts'] = df_kl_div['df_sliced_counts'].replace(0, 1e-15)  # to avoid division by zero
        df_kl_div['df_ref_counts_pct'] = [v / sum(df_kl_div['df_ref_counts'].values) for v in df_kl_div['df_ref_counts'].values]
        df_kl_div['df_sliced_counts_pct'] = [v / sum(df_kl_div['df_sliced_counts'].values) for v in df_kl_div['df_sliced_counts'].values]

        # keep in dataframe format for ease of debugging / troubleshooting
        probs_df_ref = df_kl_div['df_ref_counts_pct']
        probs_df_sliced = df_kl_div['df_sliced_counts_pct']
        return probs_df_ref, probs_df_sliced

    def _get_df_feature_with_pred_state_cls(self, df_overall: pd.DataFrame):
        '''
        For classification task only.
        Function to tap-out customized df combining features and relevant prediction info for use in visualization.
        '''
        ls_dfs_viz, _ = super().get_df_with_probability_values()

        ls_dfs_viz_featdist = []
        for df_viz in ls_dfs_viz:
            df_viz = insert_index_col(df_viz)
            df_predstate = df_viz[['index', 'yTrue', 'yPred-label', 'pred_state', 'model']]
            df_viz_interim = df_overall.merge(df_predstate, how='left', on='index')
            ls_dfs_viz_featdist.append(df_viz_interim)
        return ls_dfs_viz_featdist

    def _get_probabilities_by_feature(self, df_viz: pd.DataFrame, specific_feature: str):
        '''
        For classification task only.
        Function to calculate probabilities of correct vs miss-predict for specific feature
        '''
        df_pivot = pd.pivot_table(
            df_viz[[specific_feature, 'pred_state', 'model']],
            index=specific_feature,
            values='model',
            columns='pred_state',
            aggfunc='count',
            fill_value=1e-15)  # to avoid inf due to zero division for those NA or zero

        # set up new columns to get the percentage of each subvalue of x_colName
        df_pivot['correct'] = df_pivot['correct'].replace(0, 1e-15)  # to avoid division by zero
        df_pivot['miss-predict'] = df_pivot['miss-predict'].replace(0, 1e-15)  # to avoid division by zero
        df_pivot['correct_pct'] = [v / sum(df_pivot['correct']) for v in df_pivot['correct']]
        df_pivot['misspredict_pct'] = [v / sum(df_pivot['miss-predict']) for v in df_pivot['miss-predict']]

        probs_correct = df_pivot['correct_pct']
        probs_misspred = df_pivot['misspredict_pct']
        return probs_correct, probs_misspred

    def _generate_kl_div_info_base(self, df: pd.DataFrame, feature_to_exclude: List):
        '''
        Function to generate dictionary like output storing kl-divergence score for each feature
        arranged in descending order.
        '''
        kl_div_dict = {}
        for feat in self.features:
            if feat not in feature_to_exclude:
                if is_regression(self.analysis_type):
                    df_viz_specific_feat, optimum_bin_size = self._get_single_feature_df_with_binning(df, feat)
                    probs_df_ref, probs_df_sliced = self._get_probabilities_by_bin_group(df_viz_specific_feat, optimum_bin_size)
                    kl_div = calculate_kl_div(probs_df_ref, probs_df_sliced)
                    kl_div_dict[feat] = [kl_div, df_viz_specific_feat]

                elif is_classification(self.analysis_type):
                    df_viz_specific_feat = df[[feat, 'pred_state', 'model']]
                    try:
                        probs_correct, probs_misspred = self._get_probabilities_by_feature(df, feat)
                        kl_div = calculate_kl_div(probs_correct, probs_misspred)
                    except KeyError:  # KeyError: 'miss-predict' => no miss-predict for the selected idx range
                        kl_div = 0  # no comparison is feasible, therefore divergence is 0
                    kl_div_dict[feat] = [kl_div, df_viz_specific_feat]

        kl_div_dict_sorted = dict(sorted(kl_div_dict.items(), key=lambda x: x[1][0], reverse=True))
        return kl_div_dict_sorted

    def xform(self, feature_to_exclude: Optional[List[str]] = None, start_idx: Optional[int] = None, stop_idx: Optional[int] = None):
        '''
        Core transformation function to tap-out data into input format suitable for plotly graph

        Arguments:
            feature_to_exclude (List of :obj:`str`, `optional`):
                A list of features to be excluded from the kl-div calculation and visualization
            start_idx (:obj:`int`, `optional`):
                Integer number indicating the start index position to slice dataframe
            stop_idx (:obj:`int`, `optional`):
                Integer number indicating the stop index position to slice dataframe

        Returns:
            :obj:`Dict` or :obj:`List(Dict)`:
                dictionary storing kl-divergence score for each feature in decending order
        '''
        if isinstance(feature_to_exclude, list):
            feature_to_exclude = feature_to_exclude
        else:
            try:
                feature_to_exclude = [feature_to_exclude]
            except TypeError:
                feature_to_exclude = []

        df_sliced = self._get_df_sliced(start_idx, stop_idx)
        idx_sliced_df = list(df_sliced.index)

        df_overall = self.df_features.copy()
        df_overall['dataset_type'] = ['df_sliced' if idx in idx_sliced_df else 'df_reference' for idx in df_overall.index]

        if is_regression(self.analysis_type):
            kl_div_dict_sorted = self._generate_kl_div_info_base(df_overall, feature_to_exclude)
            kl_div_dict_sorted.pop('index')
            return kl_div_dict_sorted

        elif is_classification(self.analysis_type):
            if start_idx is not None and stop_idx is not None:  # user has specified a slicing range to inspect
                df_overall = df_overall.loc[lambda x: x['dataset_type'] == 'df_sliced', :]

            ls_dfs_viz_featdist = self._get_df_feature_with_pred_state_cls(df_overall)

            ls_kl_div_dict_sorted = []
            for df_viz in ls_dfs_viz_featdist:
                kl_div_dict_sorted = self._generate_kl_div_info_base(df_viz, feature_to_exclude)
                kl_div_dict_sorted.pop('index')
                ls_kl_div_dict_sorted.append(kl_div_dict_sorted)
        return ls_kl_div_dict_sorted



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Dict
import pandas as pd

import dash
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc

from rarity.app import app
from rarity.data_loader import CSVDataLoader, DataframeLoader
from rarity.interpreters.structured_data import IntFeatureDistribution
from rarity.visualizers.xfeature_distribution import plot_distribution_by_specific_feature, plot_distribution_by_kl_div_ranking
from rarity.utils import style_configs
from rarity.utils.common_functions import is_regression, is_classification, detected_bimodal, \
                                            invalid_slicing_range, invalid_slicing_range_sequence, invalid_min_max_limit, \
                                            incomplete_range_entry, get_max_value_on_slider


def fig_plot_distribution_by_kl_div_ranking(data_loader: Union[CSVDataLoader, DataframeLoader],
                                            feature_to_exclude: List[str],
                                            start_idx: int,
                                            stop_idx: int,
                                            display_option: str,
                                            display_value: int):
    '''
    Integration of kl-divergence scores to corresponding fig-object

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module
        feature_to_exclude (List of :obj:`str`, `optional`):
            A list of features to be excluded from the kl-div calculation and visualization
        start_idx (:obj:`int`, `optional`):
            Integer number indicating the start index position to slice dataframe
        stop_idx (:obj:`int`, `optional`):
            Integer number indicating the stop index position to slice dataframe
        display_option (str):
            - info to indicate if to display distribution plot by top-N / bottom-N or both top-N + bottom-N
            - Available options: 'top', 'bottom' or 'both'
        display_value (int):
            - number indicates the limit of graph to be displayed, max at 10
            - if dataset consists of < 10 features, the limit == no. of features the dataset has

    Returns:
        :obj:`Dict[~plotly.graph_objects.Figure]`:
            dictionary storing distribution figures by display_option

    .. note::

            if classification, returns:

                :obj:`List[Dict[~plotly.graph_objects.Figure]]`: list of dictionary storing distribution figures by display_option
    '''
    if is_regression(data_loader.analysis_type):
        kl_div_dict_sorted = IntFeatureDistribution(data_loader).xform(feature_to_exclude, start_idx, stop_idx)
        fig_obj_dict = plot_distribution_by_kl_div_ranking(kl_div_dict_sorted, display_option, display_value, 'dataset_type', None)
        return fig_obj_dict

    elif is_classification(data_loader.analysis_type):
        models = data_loader.get_model_list()
        ls_kl_div_dict_sorted = IntFeatureDistribution(data_loader).xform(feature_to_exclude, start_idx, stop_idx)
        ls_fig_obj_dict = [plot_distribution_by_kl_div_ranking(kl_div_dict, display_option, display_value, 'pred_state', models[i])
                            for i, kl_div_dict in enumerate(ls_kl_div_dict_sorted)]
        return ls_fig_obj_dict


def fig_plot_distribution_by_specific_feature(data_loader: Union[CSVDataLoader, DataframeLoader], ls_specific_feature, start_idx, stop_idx):
    '''
    Integration of kl-divergence scores to specific fig-object

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module
        ls_specific_feature (List of :obj:`str`):
            A list of features to be displayed along with the corresponding kl-div score
        start_idx (:obj:`int`, `optional`):
            Integer number indicating the start index position to slice dataframe
        stop_idx (:obj:`int`, `optional`):
            Integer number indicating the stop index position to slice dataframe

    Returns:
        :obj:`List[~plotly.graph_objects.Figure]`:
            list of figure objects displaying the distribution plot based on kl-divergence score
    '''
    if is_regression(data_loader.analysis_type):
        kl_div_dict_sorted = IntFeatureDistribution(data_loader).xform(None, start_idx, stop_idx)
        fig_objs_specific = plot_distribution_by_specific_feature(ls_specific_feature, kl_div_dict_sorted, 'dataset_type', None)
        return fig_objs_specific

    elif is_classification(data_loader.analysis_type):
        models = data_loader.get_model_list()
        ls_kl_div_dict_sorted = IntFeatureDistribution(data_loader).xform(None, start_idx, stop_idx)
        ls_fig_objs_specific = [plot_distribution_by_specific_feature(ls_specific_feature, kl_div_dict, 'pred_state', models[i])
                                for i, kl_div_dict in enumerate(ls_kl_div_dict_sorted)]
        return ls_fig_objs_specific


def _fig_layout_top_n(fig_obj_dict: Dict, display_option: int, model_pair: bool = True):
    '''
    Internal function to tapout dash html layout for top-n figures
    '''
    fig_objs_display_combined = []
    if model_pair:
        for n in range(len(fig_obj_dict[0][display_option])):
            shared_title_top_n = html.H4(f'Top {n + 1}', className='h4__title-kl-div-rank-num-top-modelpair')
            fig_objs_display_combined.append(shared_title_top_n)

            top_n_fig_cls_m1 = fig_obj_dict[0][display_option][n][0]
            top_n_fig_cls_m2 = fig_obj_dict[1][display_option][n][0]
            single_distplot_top_pair = [dbc.Row([
                                            dbc.Col([
                                                dbc.Row(dbc.Col(dcc.Graph(id=f'fig-top{n + 1}-feat-dist-cls-m1', figure=top_n_fig_cls_m1)),
                                                justify='center', className='border__kl-div-figs-both-left')], width=6),
                                            dbc.Col([
                                                dbc.Row(dbc.Col(dcc.Graph(id=f'fig-top{n + 1}-feat-dist-cls-m2', figure=top_n_fig_cls_m2)),
                                                justify='center', className='border__kl-div-figs-both-right')])
                                        ])]
            fig_objs_display_combined += single_distplot_top_pair
    else:
        for n in range(len(fig_obj_dict[display_option])):
            top_n_fig_reg = fig_obj_dict[display_option][n][0]
            single_distplot_top = [html.H4(f'Top {n + 1}', className='h4__title-kl-div-rank-num-top'),
                                    dbc.Row([
                                        dbc.Row(dcc.Graph(id=f'fig-top{n + 1}-feat-dist-reg', figure=top_n_fig_reg), justify='center')],
                                        justify='center', className='border__kl-div-figs')]
            fig_objs_display_combined += single_distplot_top
    return fig_objs_display_combined


def _fig_layout_bottom_n(fig_obj_dict: Dict, display_option: str, model_pair: bool = True):
    '''
    Internal function to tapout dash html layout for bottom-n figures
    '''
    fig_objs_display_combined = []
    if model_pair:
        for n in range(len(fig_obj_dict[0][display_option])):
            shared_title_bottom_n = html.H4(f'Bottom {n + 1}', className='h4__title-kl-div-rank-num-bottom-modelpair')
            fig_objs_display_combined.append(shared_title_bottom_n)

            btm_n_fig_cls_m1 = fig_obj_dict[0][display_option][-(n + 1)][0]  # n + 1 => to avoid -0
            btm_n_fig_cls_m2 = fig_obj_dict[1][display_option][-(n + 1)][0]  # n + 1 => to avoid -0
            single_distplot_bottom_pair = [dbc.Row([
                                            dbc.Col([
                                                dbc.Row(dbc.Col(dcc.Graph(id=f'fig-bottom{n + 1}-feat-dist-cls-m1', figure=btm_n_fig_cls_m1)),
                                                justify='center', className='border__kl-div-figs-both-left')], width=6),
                                            dbc.Col([
                                                dbc.Row(dbc.Col(dcc.Graph(id=f'fig-bottom{n + 1}-feat-dist-cls-m2', figure=btm_n_fig_cls_m2)),
                                                justify='center', className='border__kl-div-figs-both-right')])
                                        ])]
            fig_objs_display_combined += single_distplot_bottom_pair
    else:
        for n in range(len(fig_obj_dict[display_option])):
            btm_n_fig_reg = fig_obj_dict[display_option][-(n + 1)][0]  # n + 1 => to avoid -0
            single_distplot_btm = [html.H4(f'Bottom {n + 1}', className='h4__title-kl-div-rank-num-bottom'),
                                    dbc.Row([
                                        dbc.Row(dcc.Graph(id=f'fig-bottom{n + 1}-feat-dist-reg', figure=btm_n_fig_reg), justify='center')],
                                        justify='center', className='border__kl-div-figs')]
            fig_objs_display_combined += single_distplot_btm
    return fig_objs_display_combined


def _model_pair_fig_objs_layout_on_display_option(ls_fig_obj_dict: List[Dict], display_option: str):
    '''
    Internal function to collate figure layouts based on display option for bimodal scenario
    '''
    fig_objs_pairs_display_combined = []
    if display_option == 'top':
        fig_objs_pairs_display_combined = _fig_layout_top_n(ls_fig_obj_dict, display_option, model_pair=True)

    elif display_option == 'bottom':
        fig_objs_pairs_display_combined = _fig_layout_bottom_n(ls_fig_obj_dict, display_option, model_pair=True)

    elif display_option == 'both':
        fig_objs_pairs_display_combined = _fig_layout_top_n(ls_fig_obj_dict, 'top', model_pair=True) + \
                                            _fig_layout_bottom_n(ls_fig_obj_dict, 'bottom', model_pair=True)
    return fig_objs_pairs_display_combined


def _fig_objs_layout_based_on_display_option(fig_obj_dict: Dict, display_option: str):
    '''
    Internal function to collate figure layouts based on display option for non-bimodal scenario
    '''
    # fig_obj_dict => already filtered to top n value selected by user
    fig_objs_display_combined = []
    if display_option == 'top':
        fig_objs_display_combined = _fig_layout_top_n(fig_obj_dict, display_option, model_pair=False)

    elif display_option == 'bottom':
        fig_objs_display_combined = _fig_layout_bottom_n(fig_obj_dict, display_option, model_pair=False)

    elif display_option == 'both':
        fig_objs_display_combined = _fig_layout_top_n(fig_obj_dict, 'top', model_pair=False) + \
                                    _fig_layout_bottom_n(fig_obj_dict, 'bottom', model_pair=False)
    return fig_objs_display_combined


def _fig_objs_layout_based_on_specific_feature(specific_feature: List, fig_objs_specific: List):
    '''
    Internal function to collate figure layouts for specific feature/features
    '''
    fig_objs_specific_feature_combined = []
    if len(specific_feature) > 0:
        for i, specific_fig in enumerate(fig_objs_specific):
            single_displot_specific = [dbc.Row(dcc.Graph(id=f'fig-specific{i}-feat-dist-reg', figure=specific_fig), 
                                                justify='center', className='border__specific-feat-dist')]
            fig_objs_specific_feature_combined += single_displot_specific
    return fig_objs_specific_feature_combined


def _model_pair_fig_objs_layout_on_specific_feature(specific_feature: List, ls_fig_objs_specific: List):
    '''
    Internal function to collate figure layouts for specific feature/features for bimodal scenario
    '''
    fig_objs_specific_feature_combined = []
    if len(specific_feature) > 0:
        for i, (specific_fig_m1, specific_fig_m2) in enumerate(zip(ls_fig_objs_specific[0], ls_fig_objs_specific[1])):
            single_displot_specific_model_pair = [dbc.Row([
                                                        dbc.Col(dbc.Row([
                                                                    dbc.Col(dcc.Graph(id=f'fig-specific{i}-feat-dist-cls-m1',
                                                                                        figure=specific_fig_m1))],
                                                                justify='center', className='border__kl-div-figs-both-left'), width=6),
                                                        dbc.Col(dbc.Row([
                                                                    dbc.Col(dcc.Graph(id=f'fig-specific{i}-feat-dist-cls-m2',
                                                                                        figure=specific_fig_m2))],
                                                                justify='center', className='border__kl-div-figs-both-right'), width=6)])]

            fig_objs_specific_feature_combined += single_displot_specific_model_pair
    return fig_objs_specific_feature_combined


def _second_level_validity_check_on_user_input_range(range_list: List[str], df: pd.DataFrame):
    '''
    Internal function to perform validity check on user inputs on index range
    '''
    range_input_err_alert = ''
    if incomplete_range_entry(range_list):
        range_input_err_alert = style_configs.activate_incomplete_range_entry_alert()
    elif invalid_slicing_range_sequence(range_list):
        range_input_err_alert = style_configs.activate_range_input_error_alert()
    elif invalid_min_max_limit(range_list, df):
        range_input_err_alert = style_configs.activate_invalid_limit_alert(df)
    return range_input_err_alert


def _generate_valid_start_stop_idx_based_on_usr_input_range(slicing_input: str,
                                                            df_features: pd.DataFrame,
                                                            default_start_idx: int,
                                                            default_stop_idx: int):
    '''
    Internal function to adjust valid start / stop index based on user input
    '''
    if slicing_input is not None:
        if invalid_slicing_range(slicing_input):
            range_input_err_alert = style_configs.activate_range_input_error_alert()
            return '', range_input_err_alert, dash.no_update

        else:
            range_list = slicing_input.split(':')
            if '%' in slicing_input:
                if not all('%' in idx for idx in range_list):
                    range_input_err_alert = style_configs.activate_range_input_error_alert()
                    return '', range_input_err_alert, dash.no_update

                else:
                    range_input_err_obj = _second_level_validity_check_on_user_input_range(range_list, df_features)
                    if range_input_err_obj != '':
                        return '', range_input_err_obj, dash.no_update

                    start_pct = int(range_list[0].replace('%', '')) / 100
                    start_idx = int(len(df_features) * start_pct)
                    stop_pct = int(range_list[1].replace('%', '')) / 100
                    stop_idx = int(len(df_features) * stop_pct)
                    return start_idx, stop_idx
            else:
                if range_list == ['']:
                    start_idx = default_start_idx
                    stop_idx = default_stop_idx
                    return start_idx, stop_idx
                else:
                    range_input_err_obj = _second_level_validity_check_on_user_input_range(range_list, df_features)
                    if range_input_err_obj != '':
                        return '', range_input_err_obj, dash.no_update

                    start_idx, stop_idx = int(range_list[0]), int(range_list[1])
                    return start_idx, stop_idx

    elif slicing_input is None or slicing_input == '':
        start_idx = default_start_idx
        stop_idx = default_stop_idx
        return start_idx, stop_idx


class FeatureDistribution:
    '''
    Main integration for feature component on Distribution

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Important Attributes:

        analysis_type (str):
            Analysis type defined by user during initial inputs preparation via data_loader stage.
        model_names (:obj:`List[str]`):
            model names defined by user during initial inputs preparation via data_loader stage.
        is_bimodal (bool):
            to indicate if analysis involves 2 models
        feature_to_exclude (List of :obj:`str`, `optional`):
            A list of features to be excluded from the kl-div calculation and visualization
        df_features (:obj:`~pandas.DataFrame`):
            Dataframe storing all features used in dataset
        specific_feature (List of :obj:`str`):
            A list of features to be displayed along with the corresponding kl-div score
        display_option (str):
            - info to indicate if to display distribution plot by top-N / bottom-N or both top-N + bottom-N
            - Available options: 'top', 'bottom' or 'both'
        display_value (int):
            - number indicates the limit of graph to be displayed, max at 10
            - if dataset consists of < 10 features, the limit == no. of features the dataset has

    Returns:
        :obj:`~dash_core_components.Container`:
            styled dash components displaying graph and/or table objects
    '''
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader]):
        self.data_loader = data_loader
        self.analysis_type = data_loader.get_analysis_type()
        self.model_names = data_loader.get_model_list()
        self.is_bimodal = True if len(self.model_names) > 1 else False
        self.feature_to_exclude = None
        self.df_features = data_loader.get_features()
        self.specific_feature = []
        self.display_option = 'top'
        self.display_value = 3

        if is_regression(self.analysis_type):
            self.start_idx = int(len(self.df_features) * 0.8)
            self.stop_idx = len(self.df_features)
            self.range_selection_header = style_configs.DEFAULT_RANGE_SELECTION_TEXT_REG
        elif is_classification(self.analysis_type):
            self.start_idx = None
            self.stop_idx = None
            self.range_selection_header = style_configs.DEFAULT_RANGE_SELECTION_TEXT_CLS

        self.fig_obj_dict = fig_plot_distribution_by_kl_div_ranking(self.data_loader,
                                                                    self.feature_to_exclude,
                                                                    self.start_idx,
                                                                    self.stop_idx,
                                                                    self.display_option,
                                                                    self.display_value)
        self.fig_objs_specific = fig_plot_distribution_by_specific_feature(self.data_loader,
                                                                            self.specific_feature,
                                                                            self.start_idx,
                                                                            self.stop_idx)

    def show(self):
        '''
        Method to tapout styled html for feature distribution
        '''
        options_feature_ls_reg = [{'label': f'{col}', 'value': f'{col}'} for col in self.df_features.columns]

        if is_regression(self.analysis_type) or (is_classification(self.analysis_type) and not detected_bimodal(self.model_names)):
            if isinstance(self.fig_obj_dict, list):
                self.fig_obj_dict = self.fig_obj_dict[0]
            if is_classification(self.analysis_type):
                self.fig_objs_specific = self.fig_objs_specific[0]

            self.fig_objs_display_combined = _fig_objs_layout_based_on_display_option(self.fig_obj_dict, self.display_option)
            self.fig_objs_specific_feature_combined = _fig_objs_layout_based_on_specific_feature(self.specific_feature,
                                                                                                    self.fig_objs_specific)

        elif is_classification(self.analysis_type) and detected_bimodal(self.model_names):
            self.fig_objs_display_combined = _model_pair_fig_objs_layout_on_display_option(self.fig_obj_dict, self.display_option)
            self.fig_objs_specific_feature_combined = _model_pair_fig_objs_layout_on_specific_feature(self.specific_feature,
                                                                                                        self.fig_objs_specific)

        distplot = dbc.Container([
                    dbc.Row(html.H5('Feature Distribution by KL-Divergence Ranking',
                                    id='title-after-topn-bottomn-reg',
                                    className='h5__cluster-section-title')),

                    # params selection section: feature_to_exclude, select topN/BottomN/Both, define slicing range
                    html.Div([
                        dbc.Row(html.Div(html.H6('Select feature to exclude ( if applicable ) :'), className='h6__feature-to-exclude')),
                        dbc.Col(dcc.Dropdown(id='select-feature-to-exclude-featdist',
                                            options=options_feature_ls_reg,
                                        value=[], multi=True)),
                        html.Br(),
                        html.Br(),
                        dbc.Row([
                            dbc.Col([
                                dbc.Row(html.Div(
                                            html.H6('Display feature distribution by KL-Divergence score :'),
                                            className='h6__display-top-bottom-n')),
                                dbc.Row(dbc.FormGroup([
                                            dbc.RadioItems(
                                                options=[
                                                    {'label': 'Top N only', 'value': 'top'},
                                                    {'label': 'Bottom N only', 'value': 'bottom'},
                                                    {'label': 'Top N and Bottom N', 'value': 'both'}],
                                                value='top',
                                                id="select-radioitems-kldiv-top-or-bottom",
                                                inline=True, inputClassName='radiobutton__select-top-bottom-n', custom=False)])),
                                dcc.Slider(id='select-slider-top-bottom-range-featdist',
                                            min=1,
                                            max=get_max_value_on_slider(self.df_features, 'feat-dist'),  # max at 10
                                            step=1,
                                            value=3,
                                            marks=style_configs.DEFAULT_SLIDER_RANGE)], width=6),
                            dbc.Col([
                                dbc.Row(html.Div(html.H6(self.range_selection_header), className='h6__enter-data-range')),
                                dbc.Row(dbc.Input(id='input-range-to-slice-kldiv-featdist',
                                                    placeholder='start_index:stop_index',
                                                    type='text',
                                                    value=None)),
                                dbc.Row(html.Div(html.Pre(style_configs.input_range_subnote(self.df_features),
                                                            className='text__range-header-kldiv-featfist')))],
                                width=6)]),
                        html.Br(),
                        dbc.Row([
                            dbc.Col(width=10),
                            dbc.Col(dbc.Row(
                                        dcc.Loading(id='loading-output-kldiv-featdist',
                                                    type='circle', color='#a80202'),
                                    justify='right', className='loading__kldiv-featdist'), width=1),
                            dbc.Col(dbc.Row(
                                        dbc.Button("Update",
                                                    id='button-featdist-kldiv-update',
                                                    n_clicks=0,
                                                    className='button__update-dataset'),
                                    justify='right'))]),
                    ], className='border__select-dataset'),

                    html.Div(id='alert-range-input-error-kldiv-featdist'),
                    html.Div(self.fig_objs_display_combined, id='fig-objs-kl-div-by-ranking'),
                    dbc.Row(html.H5('Distribution of Specific Feature of Interest',
                                    id='title-specific-feat-dist-reg',
                                    className='h5__cluster-section-title')),

                    # params selection section: specific feature to inspect its distribution regardless of kl-div ranking
                    html.Div([
                        dbc.Row(html.Div(html.H6('Select specific feature of interest to inspect'),
                                        className='h6__cluster-instruction')),
                        dbc.Col(dcc.Dropdown(id='select-specific-feature-featdist',
                                            options=options_feature_ls_reg,
                                            value=[], multi=True)),
                        html.Br(),
                        dbc.Row([
                            dbc.Col([
                                dbc.Row(html.Div(html.H6(self.range_selection_header), className='h6__enter-data-range')),
                                dbc.Row(dbc.Input(id='input-range-to-slice-specific-feat-featdist',
                                                placeholder='start_index:stop_index',
                                                    type='text')),
                                dbc.Row(html.Div(html.Pre(style_configs.input_range_subnote(self.df_features),
                                                            className='text__range-header-kldiv-featfist')))],
                                width=6),
                            dbc.Col(width=4),
                            dbc.Col(dbc.Row(
                                        dcc.Loading(id='loading-output-specific-feat-featdist',
                                                    type='circle', color='#a80202'),
                                    justify='left', className='loading__specific-feat-featdist'), width=1),
                            dbc.Col(dbc.Row(
                                        dbc.Button("Update",
                                                    id='button-featdist-specific-feat-update',
                                                    n_clicks=0,
                                                    className='button__update-dataset'), justify='right'), )
                        ])
                    ], className='border__select-dataset'),
                    html.Div(id='alert-range-input-error-spedific-feat-featdist'),
                    html.Div(self.fig_objs_specific_feature_combined, id='fig-objs-kl-div-specific-feat'),
                    html.Br()], fluid=True)
        return distplot

    def callbacks(self):
        @app.callback(
            Output('loading-output-kldiv-featdist', 'children'),
            Output('alert-range-input-error-kldiv-featdist', 'children'),
            Output('fig-objs-kl-div-by-ranking', 'children'),
            Input('button-featdist-kldiv-update', 'n_clicks'),
            State('select-feature-to-exclude-featdist', 'value'),
            State('select-radioitems-kldiv-top-or-bottom', 'value'),
            State('select-slider-top-bottom-range-featdist', 'value'),
            State('input-range-to-slice-kldiv-featdist', 'value'))
        def update_fig_based_on_topbtm_exclusion_idx_range_params(click_count, feature_to_exclude, display_option, display_value, slicing_input):
            '''
            Callbacks functionalities on params related to top-n, botton-n / both
            '''
            if click_count > 0:
                range_input_err_alert = style_configs.no_error_alert()
                compact_outputs_kldiv = _generate_valid_start_stop_idx_based_on_usr_input_range(slicing_input,
                                                                                                self.df_features,
                                                                                                self.start_idx,
                                                                                                self.stop_idx)
                if len(compact_outputs_kldiv) == 2:
                    start_idx, stop_idx = compact_outputs_kldiv[0], compact_outputs_kldiv[1]

                    fig_obj_dict = fig_plot_distribution_by_kl_div_ranking(self.data_loader,
                                                                            feature_to_exclude,
                                                                            start_idx,
                                                                            stop_idx,
                                                                            display_option,
                                                                            display_value)
                    if is_regression(self.analysis_type) or \
                            (is_classification(self.analysis_type) and not detected_bimodal(self.model_names)):
                        if isinstance(fig_obj_dict, list):
                            fig_obj_dict = fig_obj_dict[0]
                        fig_objs_display_combined = _fig_objs_layout_based_on_display_option(fig_obj_dict, display_option)

                    elif is_classification(self.analysis_type) and detected_bimodal(self.model_names):
                        fig_objs_display_combined = _model_pair_fig_objs_layout_on_display_option(fig_obj_dict, display_option)
                    return '', range_input_err_alert, fig_objs_display_combined
                else:
                    return compact_outputs_kldiv[0], compact_outputs_kldiv[1], compact_outputs_kldiv[2]
            else:
                raise PreventUpdate

        @app.callback(
            Output('loading-output-specific-feat-featdist', 'children'),
            Output('alert-range-input-error-spedific-feat-featdist', 'children'),
            Output('fig-objs-kl-div-specific-feat', 'children'),
            Input('button-featdist-specific-feat-update', 'n_clicks'),
            State('select-specific-feature-featdist', 'value'),
            State('input-range-to-slice-specific-feat-featdist', 'value'))
        def update_fig_based_on_specific_feat_idx_range_params(click_count, specific_feature, slicing_input):
            '''
            Callbacks functionalities on params related to specific feature chosen by user
            '''
            if click_count > 0:
                range_input_err_alert = style_configs.no_error_alert()
                compact_outputs_specific_feat = _generate_valid_start_stop_idx_based_on_usr_input_range(slicing_input,
                                                                                                self.df_features,
                                                                                                self.start_idx,
                                                                                                self.stop_idx)
                if len(compact_outputs_specific_feat) == 2:
                    start_idx, stop_idx = compact_outputs_specific_feat[0], compact_outputs_specific_feat[1]
                    fig_objs_specific = fig_plot_distribution_by_specific_feature(self.data_loader, specific_feature, start_idx, stop_idx)

                    if is_regression(self.analysis_type) or \
                            (is_classification(self.analysis_type) and not detected_bimodal(self.model_names)):
                        if is_classification(self.analysis_type):
                            fig_objs_specific = fig_objs_specific[0]
                        fig_objs_specific_feature_combined = _fig_objs_layout_based_on_specific_feature(specific_feature, fig_objs_specific)

                    elif is_classification(self.analysis_type) and detected_bimodal(self.model_names):
                        fig_objs_specific_feature_combined = _model_pair_fig_objs_layout_on_specific_feature(specific_feature, fig_objs_specific)
                    return '', range_input_err_alert, fig_objs_specific_feature_combined
                else:
                    return compact_outputs_specific_feat[0], compact_outputs_specific_feat[1], compact_outputs_specific_feat[2]

            else:
                raise PreventUpdate



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union
import pandas as pd

from rarity.data_loader import CSVDataLoader, DataframeLoader
from rarity.interpreters.structured_data.base_interpreters import BaseInterpreters
from rarity.utils.common_functions import is_regression, is_classification


class IntMissPredictions(BaseInterpreters):
    '''
    Transform raw data into input format suitable for visualization on miss-prediction points

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Returns:
        :obj:`~pd.DataFrame`:
            Dataframe with essential info suitable for visualization on regression task

    .. note::

        if classification, returns:

            Compact outputs consist of the followings

            - ls_dfs_viz (:obj:`List[~pd.DataFrame]`): list of dataframes for overview visualization need
            - ls_class_labels (:obj:`List[str]`): list of class labels
            - ls_dfs_by_label  (:obj:`List[~pd.DataFrame]`): list of dataframes by individual label class
            - ls_dfs_by_label_state (:obj:`List[~pd.DataFrame]`): list of dataframes storing basic stats of each label class
    '''
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader]):
        super().__init__(data_loader)

    def xform(self):
        '''
        Core transformation function to tap-out data into input format suitable for plotly graph
        '''
        if is_regression(self.analysis_type):
            df = super().get_df_with_offset_values()
            return df

        elif is_classification(self.analysis_type):
            ls_dfs_viz, ls_class_labels = super().get_df_with_probability_values()

            ls_dfs_by_label = []
            ls_dfs_by_label_state = []
            for df_viz in ls_dfs_viz:
                # tapout df that is specific to each label-class
                dfs_specific_label = []
                dfs_state = []
                for label in ls_class_labels:
                    df_label = df_viz[df_viz['yTrue'] == int(label)]
                    df_label = df_label[['yTrue', label, 'model', 'yPred-label', 'pred_state']]
                    dfs_specific_label.append(df_label)

                    is_exist_correct = True if len(df_label[df_label['pred_state'] == 'correct']) != 0 else False
                    is_exist_misspred = True if len(df_label[df_label['pred_state'] == 'miss-predict']) != 0 else False
                    state_dict = {}
                    state_dict['sample_size: '] = len(df_label)
                    state_dict['correct: '] = len(df_label[df_label['pred_state'] == 'correct']) if is_exist_correct else 0
                    state_dict['miss-predict:'] = len(df_label[df_label['pred_state'] == 'miss-predict']) if is_exist_misspred else 0
                    state_dict['accuracy: '] = round(state_dict['correct: '] / len(df_label), 4)
                    df_state = pd.DataFrame(state_dict, index=[0],).transpose().reset_index().rename(columns={0: 'state_value'})
                    dfs_state.append(df_state)

                ls_dfs_by_label.append(dfs_specific_label)
                ls_dfs_by_label_state.append(dfs_state)
            return ls_dfs_viz, ls_class_labels, ls_dfs_by_label, ls_dfs_by_label_state



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union
import math

from rarity.data_loader import CSVDataLoader, DataframeLoader
from rarity.interpreters.structured_data.base_interpreters import BaseInterpreters
from rarity.utils.methods import create_clusters, calculate_logloss, get_optimum_num_clusters
from rarity.utils.common_functions import is_regression, is_classification


class IntLossClusterer(BaseInterpreters):
    '''
    Transform raw data into input format suitable for visualization on loss clusters

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module
    '''
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader]):
        super().__init__(data_loader)

    def extract_misspredictions(self):
        '''
        Function to tapout list of dataframe with prediction state info included
        '''
        ls_dfs_prob, _ = super().get_df_with_probability_values()
        ls_dfs_prob_misspred = [df.loc[lambda x: x['pred_state'] == 'miss-predict', :] for df in ls_dfs_prob]
        return ls_dfs_prob_misspred

    def xform(self, num_cluster: int, log_func: math.log, specific_dataset: str):
        '''
        Core transformation function to tap-out data into input format suitable for plotly graph

        Arguments:
            num_cluster (int):
                Number of cluster to form
            log_funct (:obj:`math.log`):
                Mathematics logarithm function used to calculate log-loss between yTrue and yPred
            specific_dataset (str):
                Default to 'All' indicating to include all miss-predict labels. Other options flexibly expand depending on class labels

        Returns:

                Compact outputs consist of the followings

                - df (:obj:`~pd.DataFrame`): dataframes for overview visualization need with offset values included
                - ls_score (:obj:`List[float]`): list of silhouette scores, indication of clustering quality
                - ls_cluster_range (:obj:`List[List[int]]`): list of list containing cluster number range from 1 to 10
                - ls_ssd (:obj:`List[float]`): sum of squared distance generated via kmean_inertia

        .. note::

            if classification, returns:

                Compact outputs consist of the followings

                - ls_dfs_viz (:obj:`List[~pd.DataFrame]`): dataframes for overview visualization need with offset values included
                - ls_class_labels (:obj:`List[str]`): list of all class labels
                - ls_class_labels_misspred (:obj:`List[str]`): list of class labels with minimum of 1 miss-prediction
                - ls_score (:obj:`List[float]`): list of silhouette scores, indication of clustering quality
                - ls_cluster_range (:obj:`List[List[int]]`): list of list containing cluster number range from 1 to 10
                - ls_ssd (:obj:`List[float]`): sum of squared distance generated via kmean_inertia
        '''
        if is_regression(self.analysis_type):
            df = super().get_df_with_offset_values()
            df.insert(0, 'index', df.index)  # for ease of user to trace the datapoint in raw dataframe

            cluster_groups_m1, cluster_score_m1 = create_clusters(df[f'offset_{self.models[0]}'], num_cluster)
            df[f'cluster_{self.models[0]}'] = cluster_groups_m1

            cluster_range_m1, sum_squared_distance_m1 = get_optimum_num_clusters(df[f'offset_{self.models[0]}'], num_cluster)
            # wrapped in list to std with classification output format for ease of feature integration at later stage
            ls_score = [cluster_score_m1]
            ls_cluster_range = [cluster_range_m1]
            ls_ssd = [sum_squared_distance_m1]

            if len(self.models) == 2:
                cluster_groups_m2, cluster_score_m2 = create_clusters(df[f'offset_{self.models[1]}'], num_cluster)
                df[f'cluster_{self.models[1]}'] = cluster_groups_m2

                cluster_range_m2, sum_squared_distance_m2 = get_optimum_num_clusters(df[f'offset_{self.models[1]}'], num_cluster)
                ls_score.append(cluster_score_m2)
                ls_cluster_range.append(cluster_range_m2)
                ls_ssd.append(sum_squared_distance_m2)

            return df, ls_score, ls_cluster_range, ls_ssd

        elif is_classification(self.analysis_type):
            ls_dfs_prob, ls_class_labels = super().get_df_with_probability_values()
            ls_dfs_prob_misspred = [df.loc[lambda x: x['pred_state'] == 'miss-predict', :] for df in ls_dfs_prob]
            ls_class_labels_misspred = [list(df['yTrue'].astype('str').unique()) for df in ls_dfs_prob_misspred]
            if len(ls_dfs_prob) == 2:
                ls_class_labels_misspred = list(set(ls_class_labels_misspred[0] + ls_class_labels_misspred[1]))
            else:
                ls_class_labels_misspred = ls_class_labels_misspred[0]

            try:
                ls_class_labels_misspred = [int(label) for label in ls_class_labels_misspred]
                ls_class_labels_misspred.sort()
            except TypeError:
                ls_class_labels_misspred.sort()

            ls_score = []
            ls_dfs_viz = []
            ls_cluster_range = []
            ls_ssd = []

            for df in ls_dfs_prob_misspred:
                df['eff_prob_for_loss_cal'] = [df[df['yTrue'].astype('str').values[i]].values[i] for i in range(len(df))]

                if specific_dataset != 'All':
                    try:
                        df_viz = df.loc[lambda x: x['yTrue'].astype('str') == specific_dataset, :]
                    except ValueError:
                        pass
                else:
                    df_viz = df

                df_temp = df_viz[['yTrue', 'eff_prob_for_loss_cal']]
                df_viz['lloss'] = df_temp.apply(lambda x: calculate_logloss(x['yTrue'], x['eff_prob_for_loss_cal'], log_func), axis=1)
                df_viz.pop('eff_prob_for_loss_cal')

                cluster_groups, cluster_score = create_clusters(df_viz['lloss'], num_cluster)
                df_viz['cluster'] = cluster_groups

                cluster_range, sum_squared_distance = get_optimum_num_clusters(df_viz['lloss'], num_cluster)

                ls_score.append(cluster_score)
                ls_dfs_viz.append(df_viz)
                ls_cluster_range.append(cluster_range)
                ls_ssd.append(sum_squared_distance)

            return ls_dfs_viz, ls_class_labels, ls_class_labels_misspred, ls_score, ls_cluster_range, ls_ssd



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
import pandas as pd

from sklearn.metrics import confusion_matrix, classification_report
from sklearn.metrics import roc_auc_score, roc_curve, average_precision_score, precision_recall_curve

import plotly.express as px
import plotly.graph_objects as go
import dash_table


def plot_confusion_matrix(yTrue: pd.Series, yPred: pd.Series, model_names: List):
    """
    Create confusion matrix

    Arguments:
        yTrue (:obj:`pd.Series`):
            true labels, output from int_general_metrics
        yPred (:obj:`pd.Series`):
            predicted labels, output from int_general_metrics
        model_names (:obj:`List[str]`):
            model names, output from interpreter int_general_metrics

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying confusion matrix details
    """
    fig_objs = []
    for i in range(len(yPred)):
        conf_matrix = confusion_matrix(yTrue, yPred[i], labels=list(sorted(set(yTrue))))

        fig = px.imshow(conf_matrix, 
                        labels=dict(x="Predicted Label", y="True Label", color="No.of Label"),
                        color_continuous_scale=px.colors.sequential.Viridis,
                        x=list(sorted(set(yPred[i]))),  # x = y
                        y=list(sorted(set(yTrue))),  # y = x
                        title=f'Confusion Matrix : <b>{model_names[i]}</b>') 
        fig.update_layout(title={'y': 0.90, 
                                 'x': 0.5, 
                                 'xanchor': 'center', 
                                 'yanchor': 'top', 
                                 },
                          margin=dict(r=180),
                          xaxis={'side': 'bottom'})

        fig_objs.append(fig)
    return fig_objs


def plot_classification_report(yTrue: pd.Series, yPred: pd.Series, model_names: List):
    """
    Create classification report in table form

    Arguments:
        yTrue (:obj:`pd.Series`):
            true labels, output from int_general_metrics
        yPred (:obj:`pd.Series`):
            predicted labels, output from int_general_metrics
        model_names (:obj:`List[str]`):
            model names, output from interpreter int_general_metrics

    Returns:
        :obj:`List[~plotly.graph_objects.Figure]`:
            list of tables displaying classification report details
    """  
    fig_objs = []
    for i in range(len(model_names)):
        cls_rpt = classification_report(yTrue, yPred[i], output_dict=True)
        cls_rpt_df = pd.DataFrame(cls_rpt).transpose()

        for j, ind in enumerate(list(cls_rpt_df.index)):
            if ind == 'accuracy':
                """
                To better display the accuracy record
                """
                cls_rpt_df.iloc[j, 0:2] = ''
                cls_rpt_df.iloc[j, -1] = cls_rpt_df.iloc[j + 1, -1]

        header = [''] + cls_rpt_df.columns.tolist()
        values_cells = [cls_rpt_df.index.tolist(), 
                        [f'{i:.4f}' if i != '' else i for i in cls_rpt_df['precision']],
                        [f'{i:.4f}' if i != '' else i for i in cls_rpt_df['recall']],
                        [f'{i:.4f}' for i in cls_rpt_df['f1-score']],
                        cls_rpt_df['support'].tolist()]

        fig = go.Figure(data=[go.Table(header=dict(values=header), cells=dict(values=values_cells, height=28))])
        # height => cell height

        fig.update_layout(
            title=f'Classification Report : <b>{model_names[i]}</b>', 
            title_x=0.5, 
            autosize=True,
            margin={'b': 0, 'pad': 4})
        fig_objs.append(fig)
    return fig_objs


def plot_roc_curve(yTrue: pd.Series, yPred: pd.Series, model_names: List):
    """
    Display roc curve for comparison on various models

    Arguments:
        yTrue (:obj:`pd.Series`):
            true labels, output from int_general_metrics
        yPred (:obj:`pd.Series`):
            predicted labels, output from int_general_metrics
        model_names (:obj:`List[str]`):
            model names, output from interpreter int_general_metrics

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying line curves comparing roc-auc score for various models
    """
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=[0, 1], y=[0, 1], mode='lines', line=dict(color='navy', dash='dash'), showlegend=False))

    is_multiclass = len(set(yTrue)) > 2
    if is_multiclass:
        for i in range(len(model_names)):
            fpr_j_list = []
            tpr_j_list = []
            score_j_list = []
            for j in sorted(set(yTrue)):
                yTrue_j = [1 if x == j else 0 for x in yTrue]
                yPred_j = [pred[1] if pred[0] == j else 1 - pred[1] for pred in yPred[i]]

                fpr_j, tpr_j, _ = roc_curve(yTrue_j, yPred_j)
                score_j = roc_auc_score(yTrue_j, yPred_j)
                fpr_j_list.append(fpr_j)
                tpr_j_list.append(tpr_j)
                score_j_list.append(score_j)

            for k in range(len(fpr_j_list)):
                """
                To plot roc_curve on same figure for multiple models
                """
                fig.add_trace(go.Scatter(
                    x=fpr_j_list[k], 
                    y=tpr_j_list[k], 
                    mode='lines',
                    name=f'model_{model_names[i]}_class_{list(sorted(set(yTrue)))[k]} [score: {score_j_list[k]:.4f}]', 
                    hoverlabel=dict(namelength=-1)))
    else:
        fpr_list = []
        tpr_list = []
        score_list = []
        yTrue = [int(i) for i in yTrue]
        for i in range(len(yPred)):
            fpr, tpr, _ = roc_curve(yTrue, yPred[i])
            score = roc_auc_score(yTrue, yPred[i])
            fpr_list.append(fpr)
            tpr_list.append(tpr)
            score_list.append(score)

        for i in range(len(yPred)):
            """
            To plot roc_curve on same figure for multiple models
            """
            fig.add_trace(go.Scatter(
                x=fpr_list[i], 
                y=tpr_list[i], 
                mode='lines', 
                name=f'{model_names[i]} [score: {score_list[i]:.4f}]', 
                hoverlabel=dict(namelength=-1)))

    fig.update_layout(title='<b>Receiver Operating Characteristic (ROC)</b>', 
                      xaxis_title='False Positive Rate', 
                      yaxis_title='True Positive Rate', 
                      title_x=0.2, 
                      width=500)
    return fig


def plot_precisionRecall_curve(yTrue: pd.Series, yPred: pd.Series, model_names: List):
    """
    Display precision-recall curve for comparison on various models

    Arguments:
        yTrue (:obj:`pd.Series`):
            true labels, output from int_general_metrics
        yPred (:obj:`pd.Series`):
            predicted labels, output from int_general_metrics
        model_names (:obj:`List[str]`):
            model names, output from interpreter int_general_metrics

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying line curves comparing precision-recall for various models
    """
    fig = go.Figure()

    is_multiclass = len(set(yTrue)) > 2
    if is_multiclass:
        for i in range(len(model_names)):
            precision_j_list = []
            recall_j_list = []
            score_j_list = []
            for j in sorted(set(yTrue)):
                yTrue_j = [1 if x == j else 0 for x in yTrue]
                yPred_j = [pred[1] if pred[0] == j else 1 - pred[1] for pred in yPred[i]]

                precision, recall, _ = precision_recall_curve(yTrue_j, yPred_j)
                score = average_precision_score(yTrue_j, yPred_j)
                precision_j_list.append(precision)
                recall_j_list.append(recall)
                score_j_list.append(score)

            for k in range(len(precision_j_list)):
                """
                To plot curves on same figure for multiple models
                """
                fig.add_trace(go.Scatter(
                    x=recall_j_list[k], 
                    y=precision_j_list[k], 
                    fill='tozeroy', 
                    name=f'model_{model_names[i]}_class{list(sorted(set(yTrue)))[k]} [score: {score_j_list[k]:.4f}]', 
                    hoverlabel=dict(namelength=-1)))
    else:
        precision_list = []
        recall_list = []
        score_list = []
        yTrue = [int(i) for i in yTrue]
        for i in range(len(yPred)):
            precision, recall, _ = precision_recall_curve(yTrue, yPred[i])
            score = average_precision_score(yTrue, yPred[i])
            precision_list.append(precision)
            recall_list.append(recall)
            score_list.append(score)

        for i in range(len(yPred)):
            """
            To plot curves on same figure for multiple models
            """
            fig.add_trace(go.Scatter(
                x=recall_list[i], 
                y=precision_list[i], 
                fill='tozeroy', 
                name=f'model_{model_names[i]} [score: {score_list[i]:.4f}]', 
                hoverlabel=dict(namelength=-1)))

    fig.update_layout(title='<b>Precision Recall Curve</b>', 
                      xaxis_title='Recall', 
                      yaxis_title='Precision', 
                      title_x=0.4,
                      showlegend=True)
    return fig


def plot_prediction_vs_actual(df: pd.DataFrame):
    '''
    Display scatter plot for comparison on actual values vs prediction values

    Arguments:
        df (:obj:`pd.DataFrame`):
            dataframe containing yTrue and yPred values, output from int_general_metrics

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying scatter plot comparing actual values vs prediction values
    '''
    def _modify_legend_name(fig, legend_name_dict):
        for i, dt in enumerate(fig.data):
            for element in dt:
                if element == 'name':
                    fig.data[i].name = legend_name_dict[fig.data[i].name]
        return fig

    pred_cols = [col for col in df.columns if 'yPred_' in col]
    corrected_legend_names = [col.replace('yPred_', '') for col in pred_cols]
    legend_name_dict = dict(zip(pred_cols, corrected_legend_names))

    fig = px.scatter(df, x='yTrue', y=pred_cols,
                    trendline='ols',
                    marginal_x='histogram',
                    marginal_y='histogram',
                    color_discrete_sequence=px.colors.qualitative.D3)
    fig.update_layout(
        title='<b>Comparison of Prediction (yPred) vs Actual (yTrue)</b>',
        title_x=0.12,
        xaxis_title="Actual",
        yaxis_title="Prediction",
        legend_title="", 
        legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1), 
        margin=dict(t=110, l=30, r=30))

    # scatter plot for pred_cols[0]
    hv_template0_str = f'<b>{legend_name_dict[pred_cols[0]]}</b>'
    fig.data[0].hovertemplate = hv_template0_str + '<br><br>Actual : %{x}<br>Prediction : %{y}<extra></extra>'
    # histogram-x
    fig.data[1].hovertemplate = fig.data[1].hovertemplate.replace('value', 'yTrue').replace('variable=yPred_lasso', '')
    # histogram-y
    fig.data[2].hovertemplate = fig.data[2].hovertemplate.replace('value', 'yPred').replace('variable=yPred_lasso', '')
    # trendline for scatter plot data[0]
    fig.data[3].hovertemplate = fig.data[3].hovertemplate.replace('value', 'yPred').replace('variable=yPred_', '')

    if len(pred_cols) > 1:  # Bimodal
        # scatter plot for pred_cols[1]
        hv_template2_str = f'{legend_name_dict[pred_cols[1]]}'
        fig.data[4].hovertemplate = hv_template2_str + '<br><br>Actual : %{x}<br>Prediction : %{y}<extra></extra>'
        # histogram-x
        text_to_replace = 'variable=yPred_random_forest'
        fig.data[5].hovertemplate = fig.data[5].hovertemplate.replace('value', 'yTrue').replace(text_to_replace, '')
        # histogram-y
        fig.data[6].hovertemplate = fig.data[6].hovertemplate.replace('value', 'yPred').replace(text_to_replace, '')
        # trendline for scatter plot data[1]
        fig.data[7].hovertemplate = fig.data[7].hovertemplate.replace('value', 'yPred').replace('variable=yPred_', '')

    fig = _modify_legend_name(fig=fig, legend_name_dict=dict(zip(pred_cols, corrected_legend_names)))
    return fig


def plot_prediction_offset_overview(df: pd.DataFrame):
    '''
    Display scatter plot for overview on prediction offset values

    Arguments:
        df (:obj:`~pd.DataFrame`):
            dataframe containing yTrue and yPred values, output from int_general_metrics

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying scatter plot outlining overview on prediction offset values
    '''
    pred_cols = [col for col in df.columns if 'yPred_' in col]
    corrected_legend_names = [col.replace('yPred_', '') for col in pred_cols]
    legend_name_dict = dict(zip(pred_cols, corrected_legend_names))
    max_range = int(df[pred_cols].max().max())

    offset_cols = []
    for col in pred_cols:
        offset_col = f'offset_{legend_name_dict[col]}'
        df[offset_col] = df[col] - df['yTrue']
        offset_cols.append(offset_col)

    fig = px.scatter(df, x=pred_cols[0], y=offset_cols[0], color_discrete_sequence=px.colors.qualitative.D3)
    fig.data[0].name = corrected_legend_names[0]
    fig.update_traces(showlegend=True, hovertemplate="Prediction : %{x}<br>Offset : %{y}")

    if len(pred_cols) > 1:  # Bimodal
        fig.add_trace(go.Scatter(
            x=df[pred_cols[1]], 
            y=df[offset_cols[1]], 
            name=corrected_legend_names[1], 
            mode='markers',
            marker=dict(color='#FF7F0E'),
            hovertemplate="Prediction : %{x}<br>Offset : %{y}"))

    # add reference baseline [mainly to have baseline included in legend]
    fig.add_trace(go.Scatter(
        x=[0, max_range], 
        y=[0] * 2, 
        name="Baseline [Prediction - Actual]", 
        visible=True, 
        hoverinfo='skip',
        mode='lines',
        line=dict(color="green", dash="dot")))
    # referece baseline [mainly for the dotted line in graph, but no legend generated]
    fig.add_hline(y=0, line_dash="dot")

    fig.update_layout(
                title='<b>Prediction Offset Overview</b>', 
                xaxis_title='Prediction', 
                yaxis_title='Offset from baseline', 
                title_x=0.3,
                legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1), 
                margin=dict(t=110, l=30, r=30))

    return fig


def plot_std_error_metrics(df: pd.DataFrame):
    '''
    Display table comparing various standard metrics for regression task

    Arguments:
        df (:obj:`~pd.DataFrame`):
            dataframe containing info on error metrics, output from int_general_metrics

    Returns:
        :obj:`~dash_table.DataTable`:
            table object comparing various standard metrics for regression task
    '''
    fig = dash_table.DataTable(
        id='table', 
        columns=[{'id': c, 'name': c} if c != 'Formula' else {'id': c, 'name': c, 'presentation': 'markdown'} for c in df.columns], 
        style_cell={'textAlign': 'center', 'border': '1px solid rgb(229, 211, 197)', 'font-family': 'Arial', 'margin-bottom': '0'},
        style_header={'fontWeight': 'bold', 'color': 'white', 'backgroundColor': '#7e746d ', 'border': '1px solid rgb(229, 211, 197)'},
        style_table={'width': '98%', 'margin': 'auto'},
        data=df.to_dict('records'))
    return fig



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc

from rarity.data_loader import CSVDataLoader, DataframeLoader
from rarity.interpreters.structured_data import IntGeneralMetrics
from rarity.visualizers import general_metrics as viz_general
from rarity.utils.common_functions import is_regression, is_classification


def fig_confusion_matrix(data_loader: Union[CSVDataLoader, DataframeLoader]):
    """
    Create confusion matrix

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying confusion matrix details
    """
    yTrue, yPred, model_names = IntGeneralMetrics(data_loader, 'confMat').xform()
    fig_objs = viz_general.plot_confusion_matrix(yTrue, yPred, model_names)
    return fig_objs


def fig_classification_report(data_loader: Union[CSVDataLoader, DataframeLoader]):
    """
    Create classification report in table form

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Returns:
        :obj:`List[~plotly.graph_objects.Figure]`:
            list of tables displaying classification report details
    """
    yTrue, yPred, model_names = IntGeneralMetrics(data_loader, 'classRpt').xform()
    fig_objs = viz_general.plot_classification_report(yTrue, yPred, model_names)
    return fig_objs


def fig_roc_curve(data_loader: Union[CSVDataLoader, DataframeLoader]):
    """
    Display roc curve for comparison on various models

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying line curves comparing roc-auc score for various models
    """
    yTrue, yPred, model_names = IntGeneralMetrics(data_loader, 'rocAuc').xform()
    fig_obj = viz_general.plot_roc_curve(yTrue, yPred, model_names)
    return fig_obj


def fig_precisionRecall_curve(data_loader: Union[CSVDataLoader, DataframeLoader]):
    """
    Display precision-recall curve for comparison on various models

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying line curves comparing precision-recall for various models
    """
    yTrue, yPred, model_names = IntGeneralMetrics(data_loader, 'precRecall').xform()
    fig_obj = viz_general.plot_precisionRecall_curve(yTrue, yPred, model_names)
    return fig_obj


def fig_prediction_actual_comparison(data_loader: Union[CSVDataLoader, DataframeLoader]):
    """
    Display scatter plot for comparison on actual values (yTrue) vs prediction values (yPred)

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying scatter plot comparing actual values vs prediction values
    """
    df = IntGeneralMetrics(data_loader).xform()
    fig_obj = viz_general.plot_prediction_vs_actual(df)
    return fig_obj


def fig_prediction_offset_overview(data_loader: Union[CSVDataLoader, DataframeLoader]):
    """
    Display scatter plot for overview on prediction offset values

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying scatter plot outlining overview on prediction offset values
    """
    df = IntGeneralMetrics(data_loader).xform()
    fig_obj = viz_general.plot_prediction_offset_overview(df)
    return fig_obj


def fig_standard_error_metrics(data_loader: Union[CSVDataLoader, DataframeLoader]):
    """
    Display table comparing various standard metrics for regression task

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Returns:
        :obj:`~dash_table.DataTable`:
            table object comparing various standard metrics for regression task
    """
    df = IntGeneralMetrics(data_loader, 'stdErr').xform()
    fig_obj = viz_general.plot_std_error_metrics(df)
    return fig_obj


class GeneralMetrics:
    '''
    Main integration for feature component on General Metrics.

        - On Regression: ``Prediction vs Actual``, ``Prediction vs Offset``
        - On Classification: ``Confusion Matrix``, ``Classification Report``, ``ROC``, ``Precision-Recall``

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Returns:
        :obj:`~dash_core_components.Container`:
            styled dash components displaying graph and/or table objects
    '''
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader]):
        self.data_loader = data_loader
        self.analysis_type = data_loader.get_analysis_type()

    def show(self):
        if is_regression(self.analysis_type):
            self.pred_actual = fig_prediction_actual_comparison(self.data_loader)
            self.pred_offset = fig_prediction_offset_overview(self.data_loader)
            self.std_error_metrics = fig_standard_error_metrics(self.data_loader)
            gen_metrics = dbc.Container([
                                dbc.Row([
                                    dbc.Col([
                                        dcc.Graph(figure=self.pred_actual)
                                    ], className="border__common-gen-metrics-reg-left"),
                                    dbc.Col([
                                        dcc.Graph(figure=self.pred_offset)
                                    ], className="border__common-gen-metrics-reg-right"),
                                ]), 

                                html.Div(html.Div(self.std_error_metrics, className="div__std-err")),
                                html.Br(),
                        ], fluid=True)

        elif is_classification(self.analysis_type):
            self.conf_matrix = fig_confusion_matrix(self.data_loader)
            self.cls_report = fig_classification_report(self.data_loader)
            self.roc = fig_roc_curve(self.data_loader)
            self.prec_recall = fig_precisionRecall_curve(self.data_loader)
            if len(self.conf_matrix) > 1:
                gen_metrics = dbc.Container([
                                    dbc.Row([
                                        dbc.Col([
                                            dcc.Graph(figure=self.conf_matrix[0]),
                                            dcc.Graph(figure=self.cls_report[0]),
                                        ], className="border__common-left"),

                                        dbc.Col([
                                            dcc.Graph(figure=self.conf_matrix[1]),
                                            dcc.Graph(figure=self.cls_report[1]),
                                        ], className="border__common-right")
                                    ]), 

                                    html.Div(html.Div(dcc.Graph(figure=self.roc), 
                                                                className="fig__roc-prec-recall"), 
                                                                className="border__common"),
                                    html.Div(html.Div(dcc.Graph(figure=self.prec_recall), 
                                                                className="fig__roc-prec-recall"), 
                                                                className="border__common")
                            ], fluid=True)

            elif len(self.conf_matrix) == 1:
                gen_metrics = dbc.Container([
                                    html.Div(dbc.Row([
                                        dbc.Col(dcc.Graph(figure=self.conf_matrix[0]), className="border__common-left"),
                                        dbc.Col(dcc.Graph(figure=self.cls_report[0]), className="border__common-right")
                                    ]), className="boundary__common"), 

                                    html.Div(html.Div(dcc.Graph(figure=self.roc), 
                                                                className="fig__roc-prec-recall"), 
                                                                className="border__common"),
                                    html.Div(html.Div(dcc.Graph(figure=self.prec_recall), 
                                                                className="fig__roc-prec-recall"), 
                                                                className="border__common")
                            ], fluid=True, className="boundary__common")

        return gen_metrics



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import dash
import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc
from dash.dependencies import Input, Output

from rarity.data_loader import CSVDataLoader, DataframeLoader
from rarity.features import GeneralMetrics, MissPredictions, LossClusters, FeatureDistribution, SimilaritiesCF
from rarity.app import app


class GapAnalyzer:
    '''
    GapAnalyzer is the main class object collating all developed feature components for Rarity.
    Auto-generated error analysis supports single model and bimodal (max at comparison of 2 models side by side) on tasks
    such as regression, binary classification and multiclass classification

    Args:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            This is the class object from data_loader compiling ``xfeatures``, ``yTrue``, ``yPredict`` via either ``CSVDataLoader`` \
            (for both offline and inline analysis) or ``DataframeLoader`` (for inline analysis)
        user_defined_title (str):
            Title of analysis, text field defined by user
        user_defined_port (int):
            Port defined by user to spin up Rarity's gap analysis. If not mentioned, defaults to 8000.

    Important Attributes:

        analysis_type (str):
            Analysis type defined by user. Corresponding feature components will be auto-populated based on the \
            specified analysis type. Supported analysis types : ``Regression``, ``Binary Classification``, ``Multiclass Classification``

    '''
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader], user_defined_title: str = None, user_defined_port=8000):
        self.data_loader = data_loader
        self.analysis_type = self.data_loader.get_analysis_type().replace('-', ' ').title()
        self.usr_defined_title = user_defined_title
        self.usr_defined_port = user_defined_port

    def _layout(self) -> dbc.Container:
        '''
        The main app layout of Rarity
        '''
        main_layout = dbc.Container([
                        dbc.Jumbotron(
                            dbc.Container([
                                dbc.Row([
                                    dbc.Col(html.Div([
                                        html.Img(className='header__rarity-logo', src='assets/rarity-icon.png'),
                                        html.H4('Gap Analysis with Rarity 1.0', className='header__rarity-title'),
                                        html.Pre(f'|  {self.usr_defined_title}', className='header__usr-pjt-title'),
                                    ]), width=8, md=8, sm=9, className='header__first-row-col-left align-self-center'),

                                    dbc.Col(html.Div(
                                        html.H4(self.analysis_type, className='header__analysis-type')),
                                        width='auto', className='align-self-center'),
                                ], className='justify-content-between'),

                                dbc.Row([
                                    html.Div([
                                        dcc.Tabs(id='tabs-feature-page', value='gen-metrics', children=[
                                            dcc.Tab(label='General Metrics',
                                                    value='gen-metrics',
                                                    className='header__nav-tab',
                                                    selected_className='header__nav-tab-selected',),
                                            dcc.Tab(label='Miss Predictions',
                                                    value='miss-pred',
                                                    className='header__nav-tab',
                                                    selected_className='header__nav-tab-selected',),
                                            dcc.Tab(label='Loss Clusters',
                                                    value='loss-clust',
                                                    className='header__nav-tab',
                                                    selected_className='header__nav-tab-selected'),
                                            dcc.Tab(label='xFeature Distribution',
                                                    value='xfeat-dist',
                                                    className='header__nav-tab',
                                                    selected_className='header__nav-tab-selected'),
                                            dcc.Tab(label='Similarities',
                                                    value='similarities-cf',
                                                    className='header__nav-tab',
                                                    selected_className='header__nav-tab-selected'),
                                        ], className='header__nav-row'),
                                    ]),
                                ], className='header__nav-row')
                            ], fluid=True), fluid=True, className='sticky-top'),

                        dbc.Container(html.Div(id='feature-page', children=[]), fluid=True),
                    ], fluid=True)

        return main_layout

    def run(self) -> dash.Dash:
        '''
        Spin up Rarity web application built with ``dash`` components
        '''
        app.layout = self._layout()

        @app.callback(Output('feature-page', 'children'),
                    [Input('tabs-feature-page', 'value')])
        def display_page(pathname='gen-metrics'):
            if pathname == 'gen-metrics':
                return GeneralMetrics(self.data_loader).show()
            elif pathname == 'miss-pred':
                return MissPredictions(self.data_loader).show()
            elif pathname == 'loss-clust':
                return LossClusters(self.data_loader).show()
            elif pathname == 'xfeat-dist':
                return FeatureDistribution(self.data_loader).show()
            elif pathname == 'similarities-cf':
                return SimilaritiesCF(self.data_loader).show()
            else:
                return html.Div([html.H3('feature page {}'.format(pathname))], style={'padding-left': '30px'})

        MissPredictions(self.data_loader).callbacks()
        LossClusters(self.data_loader).callbacks()
        FeatureDistribution(self.data_loader).callbacks()
        SimilaritiesCF(self.data_loader).callbacks()

        app.run_server(debug=False, port=self.usr_defined_port)



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List


class BaseLoader(ABC):
    def __init__(
        self,
        xFeatures_file: str,
        yTrue_file: str,
        yPred_file_ls: List[str] = [],
        model_names_ls: List[str] = [],
        analysis_type: str = None
    ):
        self.xFeatures = xFeatures_file
        self.yTrue = yTrue_file
        self.yPreds = yPred_file_ls
        self.models = model_names_ls
        self.analysis_type = analysis_type

    @abstractmethod
    def get_features(self):
        raise NotImplementedError()

    @abstractmethod
    def get_yTrue(self):
        raise NotImplementedError()

    @abstractmethod
    def get_yPreds(self):
        raise NotImplementedError()

    @abstractmethod
    def get_model_list(self):
        raise NotImplementedError()

    @abstractmethod
    def get_analysis_type(self):
        raise NotImplementedError()

    @abstractmethod
    def get_all(self):
        raise NotImplementedError()



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages


REQUIRED_PKGS = [
    "pandas>=1.2.4",
    "scikit-learn>=0.24.1",
    "statsmodels==0.12.2",
    "plotly>=4.14.3",
    "dash==1.20.0",
    "dash-bootstrap-components==0.12.2",
    "dash-core-components==1.16.0",
    "dash-html-components==1.1.3",
]

DOCS_EXTRAS = [
    'Sphinx >= 3.0.0',  # Force RTD to use >= 3.0.0
    'sphinx-rtd-theme==0.5.2',
    'docutils==0.16',
    'sphinxcontrib-serializinghtml==1.1.5',
    'pylons-sphinx-themes >= 1.0.8',  # Ethical Ads
    'pylons_sphinx_latesturl',
    'repoze.sphinx.autointerface',
    'sphinxcontrib-autoprogram',
]

setup(
    name='rarity',
    version='1.0.0.dev9',
    description='Data diagnostic package with minimum setup to analyze miss predictions of ML models',
    long_description=open('README.md', 'r', encoding='utf-8').read(),
    long_description_content_type="text/markdown",
    keywords='data analysis model prediction diagnostic',
    author='Yap Siew Lin @ AI Singapore, Jeanne Choo (ex-AISG), Chong Wei Yih (ex-AISG)',
    author_email='siewlin@aisingapore.org',
    license='Apache 2.0',
    url='https://github.com/aimakerspace/Rarity',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    package_data={'rarity': ['py.typed']},
    entry_points={'console_scripts': ['rarity-cli=rarity.commands.rarity_cli:main']},
    python_requires='>=3.6.0',
    install_requires=REQUIRED_PKGS,
    extras_require={'notebook': ['jupyter-client>=6.0.0', 'jupyter-core>=4.6.3', 'ipywidgets>=7.5.1'], 'docs': DOCS_EXTRAS},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
)



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Optional
import pandas as pd
from sklearn.preprocessing import StandardScaler, LabelEncoder

from rarity.data_loader import CSVDataLoader, DataframeLoader
from rarity.interpreters.structured_data.base_interpreters import BaseInterpreters
from rarity.utils.methods import compute_distances
from rarity.utils.common_functions import is_regression, is_classification, insert_index_col


class IntSimilaritiesCounterFactuals(BaseInterpreters):
    '''
    Transform raw data into input format suitable for visualization on Similarities / Counter-Factuals

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module
    '''
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader]):
        super().__init__(data_loader)
        self.df_features = data_loader.get_features()
        self.models = data_loader.get_model_list()

    def _get_categorical_features(self, df: pd.DataFrame):
        '''
        Identify categorical features
        '''
        categorical_cols = [col for col in df.columns if df[col].dtype == 'object']
        return categorical_cols

    def _label_encode_categorical_features(self, df: pd.DataFrame, categorical_cols: List):
        '''
        Fit-transform categorical features with ``LabelEncoder``
        '''
        df_encoded = df.copy()
        for col in categorical_cols:
            df_encoded[col] = LabelEncoder().fit_transform(df_encoded[col])
        return df_encoded

    def _apply_standard_scale(self, df: pd.DataFrame):
        '''
        Standard scale features
        '''
        scaler = StandardScaler()
        return scaler.fit_transform(df)

    def _get_ranking_and_distance_metrics(self, user_defined_idx: int, feature_to_exclude: List, top_n: int):
        '''
        Compute distance scores and generate index list sorted by distance ranking
        '''
        categorical_cols = self._get_categorical_features(self.df_features)
        df_encoded = self._label_encode_categorical_features(self.df_features, categorical_cols)
        scaled_data = self._apply_standard_scale(df_encoded)
        df_scaled = pd.DataFrame(scaled_data, index=df_encoded.index, columns=df_encoded.columns)
        df_scaled = df_scaled[[col for col in df_scaled.columns if col not in feature_to_exclude]]
        baseline = df_scaled.iloc[[user_defined_idx]]

        distance_metrics_dict = {}
        for idx in df_scaled.index:
            distance_metrics_dict[idx] = compute_distances(baseline, df_scaled.loc[lambda x: x.index == idx, :])[0][0]

        sorted_distance_metrics_dict = dict(sorted(distance_metrics_dict.items(), key=lambda x: x[1]))

        idx_for_top_n = [k for k in list(sorted_distance_metrics_dict.keys())[:top_n + 1]]  # top_n + 1 => first is user_defined_idx
        calculated_distance = [round(v, 4) for v in list(sorted_distance_metrics_dict.values())[:top_n + 1]]
        return idx_for_top_n, calculated_distance

    def xform(self, user_defined_idx: int, feature_to_exclude: Optional[List[str]] = None, top_n: int = 3):
        '''
        Core transformation function to tap-out data into input format suitable for plotly graph

        Arguments:
            user_defined_idx (int):
                Index of the data point of interest specified by user
            feature_to_exclude (List of :obj:`str`, `optional`):
                A list of features to be excluded from the ranking and similarities distance calculation
            top_n (int):
                Number indicating the max limit of records to be displayed based on the distance ranking

        Returns:

                Outputs consist of the followings

                - idx_for_top_n (:obj:`List[int]`): list of integer numbers indicating the ranking position in ascending order
                - calculated_distance (:obj:`List[float]`): list of calculated euclidean_distances

        .. note::

            if classification, returns:

                Outputs consist of the followings

                - df_viz (:obj:`~pd.DataFrame`): dataframes for overview visualization need with true labels and predicted labels included
                - idx_for_top_n (:obj:`List[int]`): list of integer numbers indicating the ranking position in ascending order
                - calculated_distance (:obj:`List[float]`): list of calculated euclidean_distances

        '''
        if not isinstance(feature_to_exclude, list):
            try:
                feature_to_exclude = list(feature_to_exclude)
            except TypeError:  # 'NoneType' object is not iterable
                feature_to_exclude = []

        if is_regression(self.analysis_type):
            df_viz = super().get_df_with_offset_values()
            df_viz = insert_index_col(df_viz)

            idx_for_top_n, calculated_distance = self._get_ranking_and_distance_metrics(user_defined_idx, feature_to_exclude, top_n)

        elif is_classification(self.analysis_type):
            df_features = insert_index_col(self.df_features)
            yTrue = insert_index_col(self.data_loader.get_yTrue())
            df_viz = df_features.merge(yTrue, how='left', on='index')

            yPreds = self.data_loader.get_yPreds()
            for i, model in enumerate(self.models):
                df_viz[f'yPred_{model}'] = yPreds[i]['yPred-label']

            idx_for_top_n, calculated_distance = self._get_ranking_and_distance_metrics(user_defined_idx, feature_to_exclude, top_n)
        return df_viz, idx_for_top_n, calculated_distance
